from gamegym.game import Game, Situation
from gamegym.utils import get_rng
from gamegym.distribution import Explicit
from gamegym.value_learning.valuestore import LinearValueStore
import numpy as np
import pytest
from scipy.sparse import csr_matrix


def test_init():
    LinearValueStore(shape=(3, 3))
    LinearValueStore(np.zeros((4, 3)))
    LinearValueStore(np.zeros((4, 3)), shape=(4, 3))
    with pytest.raises(Exception):
        LinearValueStore((3, 3))
    with pytest.raises(Exception):
        LinearValueStore(np.zeros((4, 3)), shape=(4, 4))


def test_value_update():
    a = np.ones((4, ))
    vs = LinearValueStore(a)
    f = [0, 2, -1, 3]
    assert vs.get(f) == pytest.approx(4.0)
    assert vs.get(np.array(f)) == pytest.approx(4.0)
    #assert vs.get(csr_matrix(f)) == pytest.approx(4.0)
    vs.update(f, -0.5)
    assert vs.values == pytest.approx([1, 0, 1.5, -0.5])
    assert vs.get(f) == pytest.approx(-3.0)


def test_norm():
    vs = LinearValueStore(shape=(2, 3), fix_mean=1.0)



#!/usr/bin/python3

from .goofspiel import Goofspiel
from .dicepoker import DicePoker
from .onecardpoker import OneCardPoker
from .matrix import MatrixGame, MatrixZeroSumGame, RockPaperScissors
from .matrix import GameOfChicken, PrisonersDilemma, MatchingPennies



from ..game import Game, Situation

# TODO: Update OCP to new Game API


class OneCardPoker(Game):
    r"""
        chance
         |
         p1
         |\
         C R
         |   \
         p2   p2
         |\   /\
         C R  C F
           |
           p1
           /\
           F C

    F - fold
    C - call
    R - raise
    """

    def __init__(self, n_cards=3):
        self.n_cards = n_cards
        card_combinations = []
        for i in range(n_cards):
            for j in range(n_cards):
                card_combinations.append((i, j))
        self.card_combinations = card_combinations
        self.card_distribution = Uniform(len(card_combinations))

    def initial_state(self):
        return OneHandPokerState(None, None, game=self)

    def players(self):
        return 2


class OneHandPokerState(Situation):

    ACTIONS1 = ("raise", "check")
    ACTIONS2 = ("call", "fold")

    def player(self):
        h = self.history
        s = len(h)
        if s == 0:
            return self.P_CHANCE
        if s == 1:
            return 0
        if s == 2:
            return 1
        if h[-1] == "raise":
            return 0
        return self.P_TERMINAL

    def actions(self):
        size = len(self.history)
        if size == 0:
            return self.game.card_distribution.values()
        if self.history[-1] == "raise":
            return self.ACTIONS2
        return self.ACTIONS1

    def chance_distribution(self):
        return self.game.card_distribution

    def values(self):
        h = self.history

        if h[-1] == "fold":
            if len(h) == 3:
                return (1, -1)
            else:
                return (-1, 1)

        if h[-1] == "call":
            bet = 2
        else:
            bet = 1

        c1, c2 = self.game.card_combinations[h[0]]
        if c1 == c2:
            return (0, 0)
        if c1 > c2:
            return (bet, -bet)
        else:
            return (-bet, bet)

    def player_information(self, player):
        pair = self.game.card_combinations[self.history[0]]
        return (pair[player], self.history[1:])



import sys
import os
import pytest

TESTS = os.path.dirname(__file__)
ROOT = os.path.dirname(TESTS)
sys.path.insert(0, os.path.join(ROOT, "gamegym"))


def pytest_configure(config):
    pytest._called_from_pytest = True


def pytest_unconfigure(config):
    del pytest._called_from_pytest


def pytest_addoption(parser):
    parser.addoption("--runslow", action="store_true", default=False, help="run slow tests")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)



from ..game import Game, Situation, ActivePlayer
from ..utils import uniform
from ..server.ui import CardBuilder, Screen

from typing import Any, Tuple
import enum
import numpy as np


class Goofspiel(Game):
    EPS = 1e-6

    class Scoring(enum.Enum):
        WINLOSS = 0
        ZEROSUM = 1
        ABSOLUTE = 2

    def __init__(self, cards: int, scoring=None, rewards=None):
        assert cards >= 1
        self.cards = cards
        self.custom_rewards = rewards is not None
        if rewards is None:
            rewards = range(1, self.cards + 1)
        self.rewards = np.array(rewards, dtype=float)
        assert len(self.rewards) == self.cards
        self.scoring = self.Scoring.WINLOSS if scoring is None else scoring
        self.players = 2

    def initial_state(self) -> Tuple[Any, ActivePlayer]:
        """
        Return the initial internal state and active player.
        """
        cset = list(range(1, self.cards + 1))
        return (([tuple(cset)] * 3, (0.0, 0.0)), ActivePlayer.new_chance(None, tuple(cset)))

    def update_state(self, state: Situation, action: Any) -> Tuple[Any, ActivePlayer, tuple]:
        """
        Return the updated internal state, active player and per-player observations.
        """
        csets, scores = state.state
        player = (len(state) - 1) % 3  # players=0,1 chance=2
        new_csets = list(csets)
        nst = list(csets[player])
        nst.remove(action)
        new_csets[player] = tuple(nst)

        # First player just bid `action`
        if player == 0:
            return ((new_csets, scores), ActivePlayer.new_player(1, new_csets[1]), ())

        # Chance just drew the prize number `action`
        if player == 2:
            return ((new_csets, scores), ActivePlayer.new_player(0, new_csets[0]), (action, ) * 3)

        # Otherwise, the second player just bid `action`
        prize = self.rewards[state.history[-2] - 1]
        first_action = state.history[-1]
        if first_action > action:
            new_obs = (1, -1, 1)
            new_scores = (scores[0] + prize, scores[1])
        elif first_action < action:
            new_obs = (-1, 1, -1)
            new_scores = (scores[0], scores[1] + prize)
        else:
            new_obs = (0, 0, 0)
            new_scores = scores

        # If fhis was not the last turn
        if len(state) + 1 < self.cards * 3:
            return ((new_csets, new_scores), ActivePlayer.new_chance(None, new_csets[2]), new_obs)

        # This was the last turn
        assert len(state) + 1 == self.cards * 3
        if self.scoring == self.Scoring.WINLOSS:
            if new_scores[0] - new_scores[1] > self.EPS:
                tscore = (1.0, -1.0)
            elif new_scores[0] - new_scores[1] < -self.EPS:
                tscore = (-1.0, 1.0)
            else:
                tscore = (0.0, 0.0)
        elif self.scoring == self.Scoring.ZEROSUM:
            tscore = (new_scores[0] - new_scores[1], new_scores[1] - new_scores[0])
        else:
            tscore = new_scores
        return ((new_csets, new_scores), ActivePlayer.new_terminal(tscore), new_obs)

    def __repr__(self):
        return "<Goofspiel({}, {}{})>".format(
            self.cards, self.scoring.name,
            ", {}".format(self.rewards) if self.custom_rewards else "")


def goofspiel_feaures_cards(state, sparse=False):
    """
    Goofspiel final state features for card value learning.

    Return a np.array containing, for every card:
    * 1 if player 0 won it
    * -1 if player 1 won it
    * 0 otherwise

    For nonterminal state, returns zero array of the appropriate size.
    """
    assert not sparse
    features = np.zeros(state.game.n_cards, dtype=np.float32)
    if state.is_terminal():
        card_seq = state.played_cards(-1)
        winners = state.winners()
        for i in range(len(features)):
            if winners[i] == 0:
                features[card_seq[i]] = 1.0
            elif winners[i] == 1:
                features[card_seq[i]] = -1.0
    return features



from ..game import Game, Situation
from ..utils import get_rng
import numpy as np

# TODO: Update this to new Game API


class LinearValueStore:
    """
    A store for an `np.array` of parameters.
    
    A value is a linear combination of features and parameters.
    Update is performed along a (possibly sparse) gradient.
    If `fixed_mean` is set, it is enforced on every update (which may be slow for large shapes).
    Optional on-demand mean, L1 and L2 regularizations.
    """

    def __init__(self,
                 initializer=None,
                 *,
                 shape=None,
                 dtype='f',
                 fix_mean=None,
                 regularize_mean=None,
                 regularize_l1=None,
                 regularize_l2=None):
        """
        The initial values are either taken from `initializer` (incl. dtype)
        or set to regularize_mean, or force_mean, or 0.0.
        """
        if initializer is not None:
            assert isinstance(initializer, np.ndarray)
            assert shape is None or tuple(shape) == initializer.shape
            self.values = initializer.copy()
        else:
            self.values = np.full(shape, fix_mean or regularize_mean or 0.0, dtype=dtype)
        self.regularize_mean = regularize_mean
        self.regularize_l1 = regularize_l1
        self.regularize_l2 = regularize_l2
        self.fix_mean = fix_mean
        if self.fix_mean is not None:
            self.values += (self.fix_mean - np.mean(self.values))

    def get(self, features):
        """
        Return the value corresponding to features (which may be a sparse array).
        """
        return np.tensordot(features, self.values, axes=len(self.values.shape))[()]

    def update(self, features, gradient):
        """
        Update the values corresponding to features by the gradien (which should be scalar).

        Always normalizes parameter mean if `fix_mean` is set.
        """
        self.values += np.multiply(features, gradient)
        if self.fix_mean is not None:
            self.values += (self.fix_mean - np.mean(self.values))

    def regularize(self, step=1e-3):
        """
        Applies gradual normalization of the mean, L1 and L2 norms (if any are set)
        with the given step size.
        """
        # normalize the mean - by step size
        if self.regularize_mean is not None:
            self.values += (self.regularize_mean - np.mean(self.values)) * step

        # TODO(gavento): additive vs multiplicative norm updates

        # normalize the L1 norm - by step size
        if self.regularize_l1 is not None:
            l1 = np.sum(np.abs(self.values))
            self.values *= 1.0 + (self.regularize_l1 - l1) * step

        # normalize the L2 norm - by step size
        if self.regularize_l2 is not None:
            l2 = np.linalg.norm(self.values)
            self.values *= 1.0 + (self.regularize_l2 - l2) * step



from ..game import Game, Situation, ActivePlayer
from ..utils import uniform


class DicePoker(Game):
    r"""
        chance
         |
         p1
        /| \
       F C  R
         |   \
         p2   p2
        /|\   /\
       F C R  F C
           |
           p1
           /\
           F C

    F - fold
    C - continue
    R - raise (double value)
    """
    ACTIONS1 = ("continue", "raise", "fold")
    ACTIONS2 = ("continue", "fold")

    def __init__(self, dice_size=6, fold_cost=2):
        self.dice_size = dice_size
        self.fold_cost = fold_cost
        dice_combinations = []
        for i in range(dice_size):
            for j in range(dice_size):
                dice_combinations.append((i, j))
        self.dice_combinations = dice_combinations
        self.dice_distribution = uniform(len(self.dice_combinations))
        self.players = 2

    def initial_state(self):
        return ((), ActivePlayer.new_chance(self.dice_distribution, self.dice_combinations))

    def _player(self, state, action):
        h = state.history
        s = len(h)
        if action == "fold":
            return ActivePlayer.TERMINAL
        if s == 0:
            return 0
        if s == 1:
            return 1
        if action == "raise" and h[-1] == "continue":
            return 0
        return ActivePlayer.TERMINAL

    def update_state(self, state, action):
        h = state.history
        s = len(h)
        pair = h[0] if h else action

        if action == "raise":
            actions = self.ACTIONS2
        else:
            actions = self.ACTIONS1
        player = self._player(state, action)

        if player >= 0:
            if s == 0:
                obs = (action[0], action[1], None)
            else:
                obs = (action, ) * 3
            return ((), ActivePlayer.new_player(player, actions), obs)

        if action == "fold":
            v = self.fold_cost
            if len(h) != 2:
                v = -v
        else:
            v = pair[0] - pair[1]

        if h[-1] == "raise":
            v *= 2
        return ((), ActivePlayer.new_terminal((v, -v)), (action, ) * 3)

    def __repr__(self):
        return "<Dicepoker({}, {})>".format(self.dice_size, self.fold_cost)



import numpy as np
import pytest
import os

from gamegym.games import MatchingPennies, RockPaperScissors, MatrixZeroSumGame, Goofspiel
from gamegym.algorithms import BestResponse, OutcomeMCCFR, exploitability, RegretStrategy, approx_exploitability
from gamegym.strategy import UniformStrategy


def test_regret():
    mc = RegretStrategy()
    rs = mc.regret_matching(np.array([-1.0, 0.0, 1.0, 2.0]))
    assert rs == pytest.approx([0.0, 0.0, 1.0 / 3, 2.0 / 3])


def test_pennies():
    np.set_printoptions(precision=3)
    g = MatchingPennies()
    mc = OutcomeMCCFR(g, seed=12)
    mcs = mc.strategies
    mc.compute(500)
    s = g.start()
    assert mcs[0].strategy((), 2) == pytest.approx([0.5, 0.5], abs=0.1)
    assert mcs[1].strategy((), 2) == pytest.approx([0.5, 0.5], abs=0.1)
    s = g.play(s, index=1)
    assert mcs[0].strategy((), 2) == pytest.approx([0.5, 0.5], abs=0.1)
    assert mcs[1].strategy((), 2) == pytest.approx([0.5, 0.5], abs=0.1)


def test_mccfr_goofspiel3():
    g = Goofspiel(3, scoring=Goofspiel.Scoring.ZEROSUM)
    mc = OutcomeMCCFR(g, seed=51)
    mc.compute(600, burn=0.5)
    mcs = mc.strategies
    us = UniformStrategy()
    s1 = g.play_sequence([2])
    assert mcs[0].strategy(s1) == pytest.approx([0., 0.9, 0.], abs=0.1)
    assert g.sample_payoff(mcs, 300, seed=12)[0] == pytest.approx([0.0, 0.0], abs=0.1)
    assert g.sample_payoff((mcs[0], us), 300, seed=13)[0] == pytest.approx([1.2, -1.2], abs=0.2)
    assert exploitability(g, 0, mcs[0]) < .55
    assert exploitability(g, 1, mcs[1]) < .55


@pytest.mark.slow
def test_mccfr_goofspiel4():
    g = Goofspiel(4, scoring=Goofspiel.Scoring.ZEROSUM)
    mc = OutcomeMCCFR(g, seed=49)
    mc.compute(10000, burn=0.5)
    mcs = mc.strategies
    for p in [0, 1]:
        exp = exploitability(g, p, mcs[p])
        aexp = approx_exploitability(g, p, mcs[p], 10000, seed=31 + p)
        print(p, exp, aexp)
        assert exp == pytest.approx(0.8, abs=0.2)
        assert aexp == pytest.approx(0.7, abs=0.2)



import pytest
from gamegym.strategy import UniformStrategy, ConstStrategy
from gamegym.utils import get_rng
from gamegym.games.matrix import *


def test_base():
    gs = [
        PrisonersDilemma(),
        GameOfChicken(),
        RockPaperScissors(),
        MatchingPennies(),
        MatrixZeroSumGame([[1, 3], [3, 2], [0, 0]], [["A", "B", "C"], [0, 1]]),
        MatrixGame([[1], [2], [3]], [["A1", "A2", "A3"]]),
        MatrixGame(np.zeros([2, 4, 5, 3], dtype=np.int32)),
    ]
    for g in gs:
        s = g.start()
        assert not s.active.is_terminal()
        assert s.active.player == 0
        assert len(s.active.actions) == g.m.shape[0]
        repr(s)
        repr(g)
    g = RockPaperScissors()
    s = g.start()
    s = g.play(s, "R")
    s = g.play(s, "P")
    assert s.active.is_terminal()
    assert ((-1, 1) == s.active.payoff).all()


def test_strategies():
    g = RockPaperScissors()
    rng = get_rng(seed=41)

    s1 = [UniformStrategy(), UniformStrategy()]
    v1 = g.sample_payoff(s1, 300, rng=rng)
    assert sum(v1[0]) == pytest.approx(0.0)
    assert v1[0] == pytest.approx([0.0, 0.0], abs=0.1)

    s2 = [
        ConstStrategy((1.0, 0.0, 0.0)),
        ConstStrategy((0.5, 0.5, 0.0)),
    ]
    v2 = g.sample_payoff(s2, 300, rng=rng)
    assert sum(v2[0]) == pytest.approx(0.0)
    assert v2[0] == pytest.approx([-0.5, 0.5], abs=0.1)



from gamegym.algorithms import OutcomeMCCFR, BestResponse
from gamegym.games import OneCardPoker
import numpy as np
import pytest


def test_onecardpoker():
    g = OneCardPoker()

    s = g.initial_state()
    assert len(s.actions()) == 9

    assert s.player() == -1
    s = s.play(1)

    assert s.player() == 0
    assert set(s.actions()) == set(["check", "raise"])

    s1 = s.play("check")
    assert s1.player() == 1
    assert set(s.actions()) == set(["check", "raise"])

    s2 = s1.play("check")
    assert s2.is_terminal()
    assert s2.values() == (-1, 1)

    s3 = s2.play("raise")
    assert s3.player() == 0
    assert set(s3.actions()) == set(["call", "fold"])

    s4 = s3.play("call")
    assert s4.is_terminal()
    assert s4.values() == (-2, 2)

    s4 = s3.play("fold")
    assert s4.is_terminal()
    assert s4.values() == (-1, 1)

    s1 = s.play("raise")
    assert s1.player() == 1
    assert set(s1.actions()) == set(["call", "fold"])

    s2 = s1.play("call")
    assert s2.is_terminal()
    assert s2.values() == (-2, 2)

    s2 = s1.play("fold")
    assert s2.is_terminal()
    assert s2.values() == (1, -1)


def test_onecardpoker_mc():

    g = OneCardPoker()
    mc = OutcomeMCCFR(g, seed=56)
    mc.compute(1000)
    #print(mc.iss)
    br = BestResponse(g, 1, {0: mc})
    #print(br.value)
    assert np.mean([g.play_strategies([mc, br], seed=i)[-1].values()[0]
                    for i in range(1000)]) > -0.4



from gamegym.utils import debug_assert, get_rng, uniform, np_uniform, sample_with_p, Distribution, cached
import numpy as np
import pytest
import random
import logging


def test_rng():
    get_rng(seed=42)
    get_rng(rng=np.random.RandomState(43))
    with pytest.raises(TypeError):
        get_rng(rng=random)
    with pytest.raises(TypeError):
        get_rng(rng=random.Random(41))


def test_debug_assert():
    debug_assert(lambda: True)
    with pytest.raises(AssertionError):
        debug_assert(lambda: False)


def test_uniform():
    assert uniform(1) == (1.0, )
    assert uniform(2) == (0.5, 0.5)
    assert len(uniform(10)) == 10
    assert sum(uniform(10)) == pytest.approx(1.0)
    assert (np_uniform(2) == np.array([0.5, 0.5])).all()
    assert len(np_uniform(10)) == 10
    assert np.sum(np_uniform(10)) == pytest.approx(1.0)


def test_sample_with_p():
    for i in range(50):
        assert sample_with_p("abcd", None)[0] in "abcd"
        assert sample_with_p("abcd", None)[1] == 0.25
        assert sample_with_p(8, None)[1] == 0.125
        assert sample_with_p(1, None)[1] == 1.0
        assert sample_with_p(None, [0.1, 0.8, 0.1])[0] in (0, 1, 2)
        assert sample_with_p(None, [0.1, 0.8, 0.1])[1] in (0.1, 0.8)
        assert sample_with_p(["a", "b"], [0.2, 0.8]) in (("a", 0.2), ("b", 0.8))


def test_distribution():
    for i in range(50):
        assert Distribution("abcd", None).sample() in "abcd"
        assert Distribution("abcd", None).sample_with_p()[1] == 0.25
        assert Distribution(8, None).sample_with_p()[1] == 0.125
        assert Distribution(1, None).sample_with_p()[1] == 1.0
        assert Distribution(None, [0.1, 0.8, 0.1]).sample_with_p()[0] in (0, 1, 2)
        assert Distribution(None, [0.1, 0.8, 0.1]).sample_with_p()[1] in (0.1, 0.8)
        assert Distribution(["a", "b"], [0.2, 0.8]).sample_with_p() in (("a", 0.2), ("b", 0.8))


class Foo:
    pass


def test_cached(tmpdir, caplog):
    runs = 0

    @cached
    def compute_x(a, b):
        nonlocal runs
        runs += 1
        return a + b

    @cached(prefix="asdf")
    def compute_y(c):
        nonlocal runs
        runs += 1
        return c

    with tmpdir.as_cwd():
        assert compute_x(1, 2) == 3
        assert runs == 1
        assert compute_x(1, 2) == 3
        assert runs == 1
        assert compute_x(1, b=2) == 3
        assert runs == 2
        assert compute_x(1, 3) == 4
        assert runs == 3
        assert compute_y(3) == 3
        assert runs == 4
        assert compute_y(3) == 3
        assert runs == 4
        foo = Foo()
        with caplog.at_level(logging.WARN):
            assert compute_y(foo) == foo
        assert 'contains a pointer address' in caplog.record_tuples[-1][2]
        assert runs == 5



from ..game import Game, Situation
from ..strategy import Strategy
from ..utils import get_rng, Distribution
from ..strategy import Strategy
import numpy as np
from typing import Iterable


class InformationSetSampler:
    """
    A helper class to sample information sets and their elements (histories) with the
    probabilities proportional to given strategies.

    This algorithm traverses all possible game histories so it may be very slow for
    medium to large games. Once precomputed, sampling is very fast: O(log |Actions|).

    Read-only attributes:

    * `game`: The `Game` instance.
    * `nodes`: The number of nodes (histories) traversed.
    * `strategies`: The strategies this was computed for.
    """

    def __init__(self,
                 game: Game,
                 strategies: Iterable[Strategy],
                 for_players: Iterable[int] = None,
                 max_nodes: int = 1e6):
        """
        Compute the information sets for given game and strategies.
        Optionally, you may limit the players this is computed for and
        the number of nodes traversed.
        """
        self.game = game
        if for_players is None:
            self.players = tuple(range(self.game.players))
        else:
            self.players = tuple(for_players)
        self.strategies = strategies
        assert len(self.strategies) == game.players
        self.nodes = 0
        self.max_nodes = max_nodes

        # temporary, {player: { observation: [RecState(prev_rec_state, prev_action, p_reach)] }}
        self._tmp_infoset_history_dist = [{} for p in range(self.game.players)]
        # temporary, {player: { observation: p_reach }}
        self._tmp_infoset_dist = [{} for p in range(self.game.players)]
        # temporary, {player: p_total }
        self._tmp_player_dist = [0.0 for p in range(self.game.players)]

        # Run the trace
        self._trace(self.game.start(), 1.0, None, None)

        # distributions are pairs: (values, probs)
        def _dist(values, probs):
            if not isinstance(values, tuple):
                values = tuple(values)
            assert len(values) == len(probs)
            s = np.sum(probs)
            assert s > 1e-6
            return (values, np.array(probs) / s)

        # Finalize the sets
        # [{ observation: Distribution( (prev_rec_state, prev_action, p_reach) ) }]
        self._infoset_history_dist = [{
            obs: Distribution(support, np.fromiter((i[2] for i in support), float), norm=True)
            for obs, support in self._tmp_infoset_history_dist[p].items()
        } for p in range(self.game.players)]
        # [Distribution(observation)]
        self._infoset_dist = [
            Distribution(
                tuple(self._tmp_infoset_dist[p].keys()),
                np.fromiter(self._tmp_infoset_dist[p].values(), float),
                norm=True) for p in range(self.game.players)
        ]
        # Distribution(player)
        self._player_dist = Distribution(None, self._tmp_player_dist, norm=True)

    def _trace(self, state, p_reach, prev_rec_state, prev_action):
        "Internal recursive history tracer."
        player = state.active.player
        rec_state = (prev_rec_state, prev_action, p_reach)
        self.nodes += 1
        if self.nodes > self.max_nodes:
            raise Exception("InformationSetSampler computation reached node limit {}.".format(
                self.max_nodes))

        if player in self.players:
            obs = state.observations[player]
            p_ihd = self._tmp_infoset_history_dist[player]
            p_ihd_set = p_ihd.setdefault(obs, list())
            p_ihd_set.append(rec_state)
            p_id = self._tmp_infoset_dist[player]
            p_id[obs] = p_id.get(obs, 0.0) + p_reach
            self._tmp_player_dist[player] += p_reach

        if state.active.is_terminal():
            return
        if state.active.is_chance():
            dist = state.active.chance
        else:
            dist = self.strategies[player].strategy(state)
        assert len(dist) == len(state.active.actions)
        for a, p_a in zip(state.active.actions, dist):
            self._trace(self.game.play(state, a), p_reach * p_a, rec_state, a)

    def sample_player(self, rng=None) -> int:
        """
        Return `(player, p_sampled)`.
        """
        return self._player_dist.sample_with_p(rng)

    def player_distribution(self) -> Distribution:
        """
        Return distribution over proportional to reach prob. of their active states.

        NOTE: The distribution is explicitly precomputed so this is instantaneus.
        """
        return self._player_dist

    def sample_info(self, player=None, rng=None):
        """
        Return `(player, info, p_sampled)`.
        
        Here `p_sampled=P(info|player)`, or `p_sampled=P(info, player)` if `player` is not given.
        """
        p_sample = 1.0
        if player is None:
            player, p = self.sample_player(rng)
            p_sample *= p
        info, p = self._infoset_dist[player].sample_with_p(rng)
        p_sample *= p
        return (player, info, p_sample)

    def info_distribution(self, player):
        """
        Return distribution over player informations (information sets) of the given player.

        NOTE: The distribution is explicitly precomputed so this is instantaneus.
        """
        return self._infoset_dist[player]

    def sample_state(self, player=None, info=None, rng=None):
        """
        Return `(player, info, state, p_sampled)`.

        Here `p_sampled=P(state|info, player)`, or `p_sampled=P(state, info|player)` if `player` is not given,
        or `p_sampled=P(state, info, player)` if none are given.
        """
        p_sample = 1.0
        if player is None:
            player, p = self._player_dist.sample_with_p(rng=rng)
            p_sample *= p
        if info is None:
            info, p = self._infoset_dist[player].sample_with_p(rng=rng)
            p_sample *= p
        rec_state, p = self._infoset_history_dist[player][info].sample_with_p(rng=rng)
        p_sample *= p
        state = self._reconstruct_state(rec_state)
        return (player, info, state, p_sample)

    def state_distribution(self, player, info):
        """
        Return distribution over the `GameState`s of the given information set.

        WARNING: This constucts all the states of the information set every time
        you call it. If you only want to sample the states, use `self.sample_state`.
        """
        dist = self._infoset_history_dist[player][info]
        return Distribution([self._reconstruct_state(rec) for rec in dist.vals], dist.probs)

    def _reconstruct_state(self, rec_state):
        "Internal, reconstructs the GameState from given history."
        if rec_state[0] is None:
            return self.game.start()
        prev_state = self._reconstruct_state(rec_state[0])
        return self.game.play(prev_state, rec_state[1])



#!/usr/bin/python3

import numpy as np
import contextlib
import logging
import time
import pytest
import pickle
from typing import Iterable, Union, Any, Optional, Iterator
from attr import attrs, attrib


def uniform(n: int) -> tuple:
    """
    Return a Uniform(n) distribution as a tuple.
    """
    return (1.0 / n, ) * n


def np_uniform(n: int) -> np.ndarray:
    """
    Return a Uniform(n) distribution as numpy array.
    """
    return np.full(n, 1.0 / n)


def sample_with_p(vals: Union[int, Iterable], probs: Optional[Iterable[float]],
                  rng=None) -> (Any, float):
    """
    Sample a single value according to probabilities.

    Returns `(val, prob_of_val)`.
    `vals` may be indexable or an `int` (then samples from `range(vals)`).
    `probs` may be indexable or `None` (then samples uniformly).
    Warning: repeated values are permitted but return their inidividual probabilities (not their sum).
    """
    return Distribution(vals, probs).sample_with_p(rng)


@attrs(cmp=True, slots=True, init=False)
class Distribution:
    """
    A minimal distribution wrapper.

    `vals` may be indexable or an `int` (then samples from `range(vals)`).
    `probs` may be indexable or `None` (then samples uniformly).
    Optionally perform distribution normalization.

    Warning: repeated values are permitted but `sample_with_p` will return their
    inidividual probabilities (not their sum).
    """

    vals = attrib(type=Union[None, int, Iterable])
    probs = attrib(type=Union[None, Iterable[float]])

    def __init__(self, vals, probs, norm=False):
        assert vals is not None or probs is not None
        self.vals = len(probs) if vals is None else vals
        self.probs = probs
        if norm and self.probs is not None:
            s = np.sum(self.probs)
            assert s > 0.0
            self.probs = self.probs / s

    def sample_with_p(self, rng=None) -> (Any, float):
        """
        Sample a single value according to probabilities, return `(val, prob_of_val)`.
        """
        rng = rng or np.random.RandomState()
        assert isinstance(rng, np.random.RandomState)
        n = self.vals if isinstance(self.vals, int) else len(self.vals)
        i = rng.choice(n, p=self.probs)
        return (i if isinstance(self.vals, int) else self.vals[i],
                self.probs[i] if self.probs is not None else 1.0 / n)

    def sample(self, rng=None) -> Any:
        """
        Sample a single value according to probabilities, return `(val, prob_of_val)`.
        """
        return self.sample_with_p(rng)[0]

    def items(self) -> Iterator:
        """
        Iterator over `(val, prob)` pairs.
        """
        vs = range(self.vals) if isinstance(self.vals, int) else self.vals
        for i in range(len(vs)):
            if self.probs is None:
                yield (vs[i], 1.0 / len(vs))
            else:
                yield (vs[i], self.probs[i])


def debug_assert(cond):
    if hasattr(pytest, "_called_from_pytest"):
        assert cond()


def _open_by_ext(fname, mode, *args, **kwargs):
    if fname.endswith('.bz2'):
        import bz2
        return bz2.BZ2File(fname, mode, *args, **kwargs)
    if fname.endswith('.gz'):
        import gzip
        return gzip.GzipFile(fname, mode, *args, **kwargs)
    if fname.endswith('.xz'):
        import lzma
        return lzma.LZMAFile(fname, mode, *args, **kwargs)
    return open(fname, mode, *args, **kwargs)


def store(obj, fname, *, protocol=4):
    """
    Pickle object into file, compressing by extension (bz2, gz, xz).
    """
    with _open_by_ext(fname, "w") as f:
        pickle.dump(obj, f, protocol=protocol)


def load(fname):
    """
    Unpickle object from a file, decompressing by extension (bz2, gz, xz).
    """
    with _open_by_ext(fname, "r") as f:
        return pickle.load(f)


def cached(prefix=None, ext="xz"):
    """
    Decorator to cache the function result in a file depending on args.

    The function source is stored together with the data, source code mismatch
    triggers recomputation. Note that *all* whitespace is ignored (incl. in strings)
    and this heuristic ignores any other source changes.

    The arguments are made a part of the filename, so different
    parameters need to be distinguishable by their `str()`.
    Also, any parameters containing pointer addresses will likely not
    be recognized as equal.

    Moreover, all whitespace in arguments is replaced by spaces and arguments are
    shortened to at most 32 characters. Full stringified args still compared on load.

    The pickled object is: `(data, full_source, [str(arg), ...], {name: str(kwarg), ...})`.
    """
    from functools import wraps
    from types import FunctionType
    import inspect
    import re
    import hashlib

    def shorten(s, maxlen=32):
        if len(s) > maxlen:
            s = s[:maxlen - 3] + "..."
        return s

    def decorate(f):
        @wraps(f)
        def wf(*args, **kwargs):
            nonlocal prefix
            if prefix is None:
                prefix = "cached-" + f.__name__
            for v in tuple(args) + tuple(kwargs.values()):
                if re.search(r'at 0x[0-9a-fA-F]{1,16}>', str(v)):
                    logging.warning(
                        "Argument string {!r} likely contains a pointer address, expect future mismatches."
                        .format(v))
            # stringified shortened arguments (for filename)
            shortargs = (','.join(shorten(str(a)) for a in args)) + ',' + (','.join(
                "{}={}".format(k, shorten(str(v))) for k, v in sorted(kwargs.items())))
            shortargs = re.sub('[\t\n\r]', ' ', shortargs).strip(',')
            # stored source code and same without whitespace
            src = inspect.getsource(f)
            nows_src = re.sub('[ \t\n\r]', '', src)
            # filename
            fname = "{}-{}.{}".format(prefix, shortargs, ext)
            # stringified args and kwargs
            str_args = [str(v) for v in args]
            str_kwargs = {str(k): str(v) for k, v in kwargs.items()}
            try:
                (d, src2, str_args2, str_kwargs2) = load(fname)
                nows_src2 = re.sub('[ \t\n\r]', '', src2)
                if nows_src == nows_src2:
                    if str_args != str_args2 or str_kwargs != str_kwargs2:
                        raise Exception(
                            "Shortened args (and filenames) match but full stringified args do not"
                        )
                    logging.debug("Successfully loaded '{}'".format(fname))
                    return d
                logging.warn("Source of {} changed, recomputing".format(f.__name__))
            except FileNotFoundError:
                pass
            d = f(*args, **kwargs)
            store((d, src, str_args, str_kwargs), fname)
            logging.debug("Successfully computed and stored '{}'".format(fname))
            return d

        return wf

    # When used as decorator without a calll parentheses, prefix=f
    if isinstance(prefix, FunctionType):
        f = prefix
        prefix = None
        return decorate(f)
    return decorate


def get_rng(rng=None, seed=None):
    """
    Hepler returning nupy RandomState, either `rng`, new one based on `seed` or random one.
    """
    if rng is not None and seed is not None:
        raise ValueError("provide only one of `seed` and `rng`.")
    if seed is not None:
        rng = np.random.RandomState(seed)
    if rng is None:
        rng = np.random.RandomState()
    if not isinstance(rng, np.random.RandomState):
        raise TypeError("Provided `rng` must be instance of `np.random.RandomState`.")
    return rng



from ..game import Game, Situation
from ..utils import get_rng
import numpy as np

# TODO: Update this to new Game API


class SparseSGDLinearValueLearning:
    def __init__(self, game, feature_extractor, value_store, infosetsampler, rng=None, seed=None):
        self.rng = get_rng(rng=rng, seed=seed)
        self.game = game
        self.store = value_store
        self.feature_extractor = feature_extractor
        self.infosetsampler = infosetsampler

    def iteration(self, strategies, step=1e-3, regularize_step=1e-3, val_samples=1,
                  grad_samples=1):
        # sample a player, info and actions
        player, info, _ = self.infosetsampler.sample_info(rng=self.rng)
        _, _, s0, _ = self.infosetsampler.sample_state(player=player, info=info, rng=self.rng)
        # Sample actions to equalize
        a1 = strategies[player].distribution(s0).sample(rng=self.rng)
        a2 = strategies[player].distribution(s0).sample(rng=self.rng)
        if a1 == a2:
            return

        def sample_term_features(a):
            _, _, s, _ = self.infosetsampler.sample_state(player=player, info=info, rng=self.rng)
            s = s.play(a)
            z = self.game.play_strategies(strategies, rng=self.rng, state0=s)[-1]
            return self.feature_extractor(z)

        def sample_features(a, samples):
            return np.mean([sample_term_features(a) for i in range(samples)], axis=0)

        a1val = self.store.get(sample_features(a1, val_samples))
        a2val = self.store.get(sample_features(a2, val_samples))
        d = (a1val - a2val) * step
        self.store.update(sample_features(a1, grad_samples), -d)
        self.store.update(sample_features(a2, grad_samples), d)
        self.store.regularize(regularize_step)

    def compute(self,
                strategies,
                iterations,
                step=1e-3,
                regularize_step=1e-3,
                record_every=None,
                val_samples=1,
                grad_samples=1):
        params = []
        for i in range(iterations):
            self.iteration(
                strategies,
                step=step,
                regularize_step=regularize_step,
                val_samples=val_samples,
                grad_samples=grad_samples)
            if record_every and i % record_every == 0:
                params.append(self.store.values.copy())
        if record_every:
            return np.array(params)



import pytest
from gamegym.games import Goofspiel
from gamegym.strategy import UniformStrategy


def test_goofspiel():
    g = Goofspiel(7)
    s = g.start()

    assert s.active.is_chance()
    assert s.active.actions == tuple(range(1, 8))
    assert (s.active.chance == (pytest.approx(1 / 7), ) * 7)

    for i, a in enumerate([4, 2, 1, 5, 4, 6, 6, 3, 3, 2, 5, 4, 3, 7]):
        s = g.play(s, a)

    assert s.active.player == 1
    assert s.active.actions == (2, 5, 7)
    assert [o.obs for o in s.observations[2]] == [4, 1, 5, -1, 6, 0, 2, 1, 3]
    assert s.state[1] == pytest.approx([6, 5])
    assert s.state[0][0] == (1, 6)
    assert s.state[0][1] == (2, 5, 7)
    assert s.state[0][2] == (1, 7)

    for a in [2, 7, 6, 7, 1, 1, 5]:
        s = g.play(s, a)

    assert s.active.is_terminal()
    assert s.state[1] == pytest.approx([9, 13])
    assert s.active.payoff == pytest.approx([-1.0, 1.0])


def test_goofspiel_rewards():
    us = [UniformStrategy(), UniformStrategy()]
    g = Goofspiel(2, Goofspiel.Scoring.ZEROSUM, rewards=[100, 11])
    for i in range(50):
        s = g.play_strategies(us, seed=i)
        assert s.active.payoff in ((0.0, 0.0), (-89.0, 89.0), (89.0, -89.0))

    g = Goofspiel(2, Goofspiel.Scoring.ABSOLUTE, rewards=[100, 11])
    for i in range(50):
        s = g.play_strategies(us, seed=i)
        assert s.active.payoff in ((0.0, 0.0), (100.0, 11.0), (11.0, 100.0))



from ..strategy import DictStrategy


def write_efg(game, f, names=False):
    """
    Write out the game tree in Gambit EFG text format.

    Optionally also write names of all the nodes, actions and info-sets.
    """

    def esc(s):
        return str(s).replace('"', '\\"').replace("\n", " ")

    def escn(s):
        return esc(s) if names else ""

    # infoset: (player, obs) -> int
    infosets = {}
    # counters for unique outcomes and chance infosets
    outcomes = 0
    chance_infosets = 0

    pls = range(game.players)
    f.write('EFG 2 R "{}" {{ {} }}\n'.format(
        esc(game), ' '.join('"Player {}"'.format(p + 1) for p in pls)))

    def traverse(state):
        nonlocal outcomes, chance_infosets
        if state.active.is_terminal():
            v = state.active.payoff
            outcomes += 1
            f.write('t "{}" {} "" {{ {} }}\n'.format(
                escn(state.history), outcomes, ' '.join("{:.6f}".format(v[p]) for p in pls)))
        elif state.active.is_chance():
            chance_infosets += 1
            d = state.active.chance
            f.write('c "{}" {} "" {{ {} }} 0\n'.format(
                escn(state.history), chance_infosets, ' '.join(
                    '"{}" {:.6f}'.format(esc(a), p) for a, p in zip(state.active.actions, d))))
            for a in state.active.actions:
                traverse(game.play(state, a))
        else:
            obs = state.observations[state.active.player]
            iset = infosets.setdefault(obs, len(infosets) + 1)
            f.write('p "{}" {} {} "OBS{}" {{ {} }} 0\n'.format(
                escn(state.history), state.active.player + 1, iset, escn(obs),
                ' '.join('"{}"'.format(esc(a)) for a in state.active.actions)))
            for a in state.active.actions:
                traverse(game.play(state, a))

    traverse(game.start())


def parse_strategy(game, s):
    d = s.strip().split(",")
    if d[0].strip() not in ("end", "NE"):
        raise Exception("Unknown strategy tag {r}".format(d[0]))
    probs = [float(x) for x in d[1:]]

    # infoset: player -> (obs -> int)
    infosets = [set() for _ in range(game.players)]
    # obss: player -> infoset_no -> (obs, actions)
    obss = [[] for _ in range(game.players)]

    def traverse(state):
        if state.active.is_terminal():
            pass
        elif state.active.is_chance():
            for a in state.active.actions:
                traverse(game.play(state, a))
        else:
            p = state.active.player
            obs = state.observations[p]
            if obs not in infosets[p]:
                infosets[p].add(obs)
                obss[p].append((obs, len(state.active.actions)))
            for a in state.active.actions:
                traverse(game.play(state, a))

    traverse(game.start())

    dicts = [{} for _ in range(game.players)]
    for p in range(game.players):
        for obs, actions in obss[p]:
            ps = tuple(probs[:actions])
            if len(ps) < actions:
                raise Exception("Not enough values, got {} for observation {} ({} actions)".format(
                    ps, obs, actions))
            if abs(sum(ps) - 1.0) >= 0.01:
                raise Exception("Player {} info set {} probabilities {} do not sum to 1.0".format(
                    p, obs, ps))
            probs = probs[actions:]
            dicts[p][obs] = ps
    return [DictStrategy(d) for d in dicts]



#!/usr/bin/python3

import collections
from ..strategy import Strategy
from ..game import ActivePlayer, Situation
from ..utils import get_rng, debug_assert, uniform, np_uniform

from attr import attrs, attrib
import numpy as np
import pickle
import logging
import time
import bz2
import sys


class RegretStrategy(Strategy):
    """
    Average strategy and regret storage for one player.

    Dictionary based, any unknown observations return uniform distributions.
    """
    EPS = 1e-30

    def __init__(self):
        # observation: (regrets, strategy)
        self.table = {}
        # usage statistics
        self.queries = 0
        self.misses = 0
        # training statistics
        self.updates = 0
        self.iterations = 0

    def get_entry(self, observation: tuple, actions: int) -> tuple:
        assert isinstance(observation, tuple)
        entry = self.table.get(observation, None)
        if entry is None:
            entry = (np.zeros(actions), np.zeros(actions))
        else:
            assert len(entry[0]) == actions
            assert len(entry[1]) == actions
        return entry

    def update_entry(self, observation: tuple, actions: int, dr=None, ds=None) -> tuple:
        assert isinstance(observation, tuple)
        entry = self.table.get(observation, None)
        if entry is None:
            entry = (np.zeros(actions), np.zeros(actions))
        nr = (entry[0] + dr) if dr is not None else entry[0]
        ns = (entry[1] + ds) if ds is not None else entry[1]
        self.table[observation] = (nr, ns)

    def regret_matching(self, regret):
        "Return stratefy distribution based on the regret vector"
        regplus = np.clip(regret, 0, None)
        s = np.sum(regplus)
        if s > self.EPS:
            return regplus / s
        else:
            return np_uniform(len(regret))

    def _strategy(self, observation, n_actions: int, state: Situation = None) -> tuple:
        self.queries += 1
        entry = self.table.get(observation, None)
        if entry is not None and np.sum(entry[1]) > self.EPS:
            dist = entry[1] / np.sum(entry[1])
        else:
            dist = np_uniform(n_actions)
            self.misses += 1
        assert n_actions == len(dist)
        return dist


class MCCFRBase:
    """
    Common base for Outcome and External sampling MC CFR.
    """

    def __init__(self, game, strategies=None, update=None, seed=None, rng=None):
        self.game = game
        self.rng = get_rng(rng, seed)
        self.strategies = strategies
        if self.strategies is None:
            self.strategies = tuple(RegretStrategy() for i in range(game.players))
        assert len(self.strategies) == game.players
        self.update = update
        if self.update is None:
            self.update = tuple(range(game.players))
        for i in self.update:
            assert isinstance(self.strategies[i], RegretStrategy)
        # stats
        self.iterations = 0
        self.nodes_traversed = 0

    def compute(self, iterations, epsilon=0.6, weight=1.0, progress=True, burn=0.0):
        """
        Run MC CFR for given iterations.

        Optionally uses a progress bar (default on).

        Updates to the cummulative strategy and regret are weighted by `weight`, this
        allows you to discount early iterations.
        If `burn > 0.0`, perform smooth burn-in by multiplying `weight` by a coefficient
        going from 0.03 to 1.0 in the first `burn`-fraction of iterations (off by default,
        a sensible choice is e.g. 0.3).
        """
        log = logging.getLogger('gamegym.MCCFR')
        log.debug("Computing {} for {} (iterations={}, weight={:.4g}, epsilon={:.4g})".format(
            self.__class__.__name__, repr(self.game), iterations, weight, epsilon))
        its = range(iterations)
        if progress:
            import tqdm
            its = tqdm.tqdm(its, desc="MCCFR")
        if burn > 0.0:
            assert burn <= 1.0
        for i in its:
            self.iterations += 1
            if i < burn * iterations:
                # This was tuned on Goofspiel(4); range 0.01-0.1 seems to be reasonable.
                w = 0.03**(1.0 - float(i) / iterations / burn)
            else:
                w = 1.0
            if progress:
                r = "nodes: {}".format(self.nodes_traversed)
                if burn > 0.0:
                    r += ", burn-in: {:.2f}".format(w)
                its.set_postfix_str(r)
            for player in self.update:
                self.sampling(player, epsilon=epsilon, weight=w * weight)
            # When only one player is updated, we need to traverse as a dummy player
            # to update the cumulative strategies.
            # Using player -1 as the updated one is a bit hacky but works.
            if len(self.update) == 1:
                self.sampling(-1, epsilon=epsilon, weight=w * weight)

    def sampling(self, player, epsilon=0.6, weight=1.0):
        "Run one sampling run for the given player."
        raise NotImplementedError


class OutcomeMCCFR(MCCFRBase):
    def _outcome_sampling(self, state, player_updated, p_reach_updated, p_reach_others, p_sample,
                          epsilon, weight):
        """
        Based on Alg 3 from PhD_Thesis_MarcLanctot.pdf and cfros.cpp from his bluff11.zip.
        Returns `(utility, p_tail, p_sample_leaf)`.
        """
        self.nodes_traversed += 1

        if state.active.is_terminal():
            #print("\n### {}: player {}, history {}, payoff {}".format(
            #    self.iteration, player_updated, state.history, state.values()[player_updated]))
            return (state.active.payoff[player_updated], 1.0, p_sample)

        if state.active.is_chance():
            ai = self.rng.choice(len(state.active.actions), p=state.active.chance)
            state2 = self.game.play(state, index=ai)
            # No need to factor in the chances in Outcome sampling
            return self._outcome_sampling(state2, player_updated, p_reach_updated, p_reach_others,
                                          p_sample, epsilon, weight)

        # Extract misc, read entry from storage
        player = state.active.player
        strat = self.strategies[player]
        obs = state.observations[player]
        actions = state.active.actions

        # Treat static players as chance nodes
        if player not in self.update:
            dist = strat.strategy(state)
            ai = self.rng.choice(len(state.active.actions), p=dist)
            state2 = self.game.play(state, index=ai)
            # No need to factor in the chances in Outcome sampling
            return self._outcome_sampling(state2, player_updated, p_reach_updated, p_reach_others,
                                          p_sample, epsilon, weight)

        # Create dists, sample the action
        entry = strat.get_entry(obs, len(actions))
        dist = strat.regret_matching(entry[0])
        # exploration in self-actions
        if player == player_updated:
            dist_sample = dist * (1.0 - epsilon) + 1.0 * epsilon / len(actions)
        else:
            dist_sample = dist
        debug_assert(lambda: np.abs(np.sum(dist) - 1.0) < 1e-3)
        debug_assert(lambda: np.abs(np.sum(dist_sample) - 1.0) < 1e-3)
        action_idx = self.rng.choice(len(actions), p=dist_sample)
        action = actions[action_idx]

        # Future state
        state2 = self.game.play(state, index=action_idx)

        if player == player_updated:
            payoff, p_tail, p_sample_leaf = self._outcome_sampling(
                state2, player_updated, p_reach_updated * dist[action_idx], p_reach_others,
                p_sample * dist_sample[action_idx], epsilon, weight)
            dr = np.zeros_like(entry[0])
            U = payoff * p_reach_others / p_sample_leaf
            #print(U, payoff, p_reach_others, p_sample_leaf, p_tail, dist)
            for ai in range(len(actions)):
                if ai == action_idx:
                    dr[ai] = U * (p_tail - p_tail * dist[action_idx])
                else:
                    dr[ai] = -U * p_tail * dist[action_idx]
            strat.update_entry(obs, len(actions), dr=dr * weight)
        else:
            payoff, p_tail, p_sample_leaf = self._outcome_sampling(
                state2, player_updated, p_reach_updated, p_reach_others * dist[action_idx],
                p_sample * dist_sample[action_idx], epsilon, weight)
            strat.update_entry(
                obs, len(actions), ds=p_reach_updated / p_sample_leaf * dist * weight)

        return (payoff, p_tail * dist[action_idx], p_sample_leaf)

    def sampling(self, updated_player, epsilon=0.6, weight=1.0):
        """
        Run one outcome sampling for the given player.
        """
        s0 = self.game.start()
        self._outcome_sampling(s0, updated_player, 1.0, 1.0, 1.0, epsilon=epsilon, weight=weight)



#!/usr/bin/python3

from .bestresponse import BestResponse, exploitability, ApproxBestResponse, approx_exploitability
from .mccfr import OutcomeMCCFR, RegretStrategy
from .infosets import InformationSetSampler



#!/usr/bin/python3

import numpy as np
from typing import Union, Any

from .game import Game, Situation, ActivePlayer
from .utils import uniform


class Strategy:
    """
    Base class for a strategy.
    """

    def _strategy(self, observation: tuple, n_actions: int, state: Situation = None) -> tuple:
        """
        Action distribution in an infoset. To be implemented by individual Strategies.
        
        Return a distribution vector on action indexes.
        Wrapped by `strategy()` for checks and convenience.
        Never called for terminal states or chance nodes.

        Should not generally depend on `state`, it is provided for
        e.g. debugging and may be `None` in some situations.
        """
        raise NotImplementedError

    def strategy(self, observation_or_state: Union[Situation, tuple],
                 n_actions: int = None) -> Union[tuple, np.ndarray]:
        """
        Returns a distribution vector on action indexes for given observation or state.

        If called on an observation, the number of actions must be also provided.
        Raises `ValueError` when called for terminal states or chance nodes. 
        """
        s = observation_or_state
        if isinstance(s, Situation):
            if n_actions is not None:
                raise ValueError("Do not provide `n_action` when calling with `Situation`")
            p = s.active.player
            if p < 0:
                raise ValueError("Strategy called in non-player situation {}", s)
            d = self._strategy(s.observations[p], len(s.active.actions), s)
        elif isinstance(s, tuple):
            if n_actions is None:
                raise ValueError("Provide `n_action` when calling with observation sequence")
            assert isinstance(n_actions, int)
            d = self._strategy(s, n_actions, None)
        else:
            raise TypeError("Provide GameState or observation tuple")
        assert isinstance(d, (tuple, np.ndarray))
        return d


class UniformStrategy(Strategy):
    """
    Strategy that plays uniformly random action from those avalable.
    """

    def _strategy(self, observation: Any, n_actions: int, state: Situation = None) -> tuple:
        """
        Returns a uniform distribution on actions for the current state.
        """
        return uniform(n_actions)


class ConstStrategy(Strategy):
    """
    A strategy that always returns a single distribution.

    (Useful e.g. for matrix games.)
    """

    def __init__(self, dist):
        self.dist = dist

    def _strategy(self, observation: Any, n_actions: int, state: Situation = None) -> tuple:
        assert n_actions == len(self.dist)
        return self.dist


class DictStrategy(Strategy):
    """
    A strategy that plays according to a given dictionary.

    The dictionary has the form `observations: distribution` where `distribution`
    is a tuple or a numpy array.
    If `default_uniform` is set, uniform strategy is returned for unknown observations,
    otherwise `KeyError` is raised.
    """

    def __init__(self, dictionary: dict, default_uniform: bool = False):
        self.dictionary = dictionary
        self.default_uniform = default_uniform

    def _strategy(self, observation: Any, n_actions: int, state: Situation = None) -> tuple:
        if self.default_uniform:
            dist = self.dictionary.get(observation, None)
            if dist is None:
                dist = uniform(len(n_actions))
        else:
            dist = self.dictionary[observation]
        assert n_actions == len(dist)
        return dist



# Bees!

import numpy as np


class FeatureExtractor:
    def features(self, state):
        """
        Return features for the given state, for example:
        * For terminal nodes, this may be the numpy feature vector.

        The returned type depends on the ValueStore used, common are hashable objects
        (for tabular stores) and numpy arrays (for linear features and neural nets).
        """
        raise NotImplementedError


class HistoryFeature(FeatureExtractor):
    "Extracts the full state history."

    def features(self, state):
        "Returns `state.history`."
        return state.history


class PlayerInfoFeature(FeatureExtractor):
    "Extracts the current player's information together with the player."

    def features(self, state):
        "Returns `(active_player, her_information)`."
        p = state.player()
        return (p, state.player_information(p))


class ValueStore:
    """
    Stores and updates values (a numpy array) for some state properties:
    * Values of the regret and policy for all actions of the active player.
    * Estimated values of terminal nodes for all players.

    Store can perform regularizations, either automatically on update (e.g. ensuring zero sum)
    or when requested by `self.regularize()`.
    """

    def get_values(self, features, size=None):
        raise NotImplementedError

    def update_values(self, features, gradient):
        raise NotImplementedError

    def regularize(self, alpha):
        pass


class TabularStore(ValueStore):
    def __init__(self, dimension=None, dtype=np.float32, default=0.0):
        self.dimension = dimension
        self.dtype = dtype
        self.default = default
        self.store = {}

    def get_values(self, features, size=None):
        assert not (self.dimension is None and size is None)  # Bees!
        if features in self.store:
            return self.store[features]
        return np.zeros(
            size if size is not None else self.dimension, dtype=self.dtype) + self.default

    def update_values(self, features, gradient):
        if features in self.store:
            self.store[features] += gradient
        else:
            self.store[features] = gradient + self.default


class LinearSuccinctStore(ValueStore):
    def __init__(self, values):
        self.values = np.array(values)
        assert len(self.values.shape) == 2

    def get_values(self, features):
        return

    def update_values(self, features, gradient):
        if features in self.store:
            self.store[features] += gradient
        else:
            self.store[features] = gradient + self.default



"""
Various benchmarks.
"""

import pytest
from gamegym import games, algorithms, contrib


@pytest.mark.parametrize("gname,g,seq,reuse", [
    ("DicePoker", games.DicePoker(), [0, 1, 0], False),
    ("DicePoker", games.DicePoker(), [0, 1, 0], True),
    ("RPS", games.RockPaperScissors(), [0, 0], False),
    ("RPS", games.RockPaperScissors(), [0, 0], True),
    ("Goofspiel(4)", games.Goofspiel(4), [0] * 12, False),
    ("Goofspiel(4)", games.Goofspiel(4), [0] * 12, True),
    ("Goofspiel(6)", games.Goofspiel(6), [0] * 18, False),
    ("Goofspiel(6)", games.Goofspiel(6), [0] * 18, True),
])
def test_playthrough(benchmark, gname, g, seq, reuse):
    def bench():
        s = g.start()
        for i in seq:
            s = g.play(s, index=i, reuse=reuse)

    benchmark(bench)


@pytest.mark.parametrize("gname,g", [
    ("DicePoker", games.DicePoker()),
    ("RPS", games.RockPaperScissors()),
    ("Goofspiel(4)", games.Goofspiel(4)),
    ("Goofspiel(6)", games.Goofspiel(6)),
])
def test_outer_mccfr_sample(benchmark, gname, g):
    mc = algorithms.OutcomeMCCFR(g, seed=52)
    # small warmup
    mc.compute(64)

    def bench():
        mc.sampling(0, 0.6, 1.0)
        mc.sampling(1, 0.6, 1.0)

    benchmark(bench)



from gamegym.games import RockPaperScissors, Goofspiel
from gamegym.strategy import ConstStrategy, UniformStrategy
from gamegym.algorithms import BestResponse, ApproxBestResponse
import pytest


def test_best_response_rps():
    bart_simpson_strategy = ConstStrategy((1, 0, 0))
    game = RockPaperScissors()
    for p in [0, 1]:
        strategy = BestResponse(game, p, [bart_simpson_strategy] * 2)
        assert tuple(strategy.best_responses.values())[0] == pytest.approx((0.0, 1.0, 0.0))
        assert strategy.value == pytest.approx(1.0)


def test_approx_best_response_rps():
    bart_simpson_strategy = ConstStrategy((1, 0, 0))
    game = RockPaperScissors()

    for p in [0, 1]:
        s = ApproxBestResponse(game, 0, [bart_simpson_strategy] * 2, iterations=200, seed=23)
        assert s.strategy((), 3) == pytest.approx((0.0, 1.0, 0.0))
        assert s.sample_value(50) == pytest.approx(1.0)


@pytest.mark.slow
def test_best_response_goofspiel():
    for n_cards, br_value in [(3, pytest.approx(4 / 3)), (4, pytest.approx(2.5))]:
        game = Goofspiel(n_cards, Goofspiel.Scoring.ZEROSUM)
        strategy = BestResponse(game, 0, [UniformStrategy()] * 2)
        for k, v in strategy.best_responses.items():
            reward = k[-1].obs
            played_cards = [o.obs for o in k[0::3]]
            idx = len([i for i in range(n_cards) if i < reward and i not in played_cards])
            assert reward in played_cards or v[idx] == 1.0
        assert strategy.value == br_value


@pytest.mark.slow
def test_approx_best_response_goofspiel():
    for n_cards, its, br_value in [(3, 1000, 1.333), (4, 20000, 2.5)]:
        game = Goofspiel(n_cards, Goofspiel.Scoring.ZEROSUM)
        strategy = ApproxBestResponse(game, 0, [UniformStrategy()] * 2, iterations=its, seed=35)
        assert strategy.sample_value(its // 2) == pytest.approx(br_value, rel=0.2)


def test_best_response_limit():
    game = Goofspiel(3)
    BestResponse(game, 0, [UniformStrategy()] * 2)
    with pytest.raises(Exception, message="traversed more than"):
        BestResponse(game, 0, [UniformStrategy()] * 2, max_nodes=1024)



#!/usr/bin/python3

import collections
import attr
import numpy as np
from typing import List, Tuple, Optional, Hashable, Callable, Any, Union

from .utils import debug_assert, get_rng, uniform


@attr.s(slots=True, cmp=True, frozen=True, repr=False)
class Observation:
    """
    Single piece of new observation.
    
    Represents either the action of the active player or other observation.
    """
    OBSERVATION = 1
    OWN_ACTION = 2

    kind = attr.ib(type=int)
    obs = attr.ib(type=Hashable)

    def __repr__(self):
        if self.kind == self.OBSERVATION:
            return "Obs({})".format(self.obs)
        return "Own({})".format(self.obs)


@attr.s(slots=True, cmp=False, frozen=True)
class ActivePlayer:
    """
    Game mode description: active player, actions, payoffs (in terminals), chance distribution.
    """
    CHANCE = -1
    TERMINAL = -2

    player = attr.ib(type=int)
    actions = attr.ib(type=tuple)
    payoff = attr.ib(type=Union[tuple, np.ndarray])
    chance = attr.ib(type=Union[tuple, np.ndarray])

    @classmethod
    def new_player(cls, p, actions):
        assert p >= 0
        return cls(p, actions, None, None)

    @classmethod
    def new_chance(cls, chance, actions):
        if chance is None:
            chance = uniform(len(actions))
        assert len(actions) == len(chance)
        assert len(actions) >= 0
        return cls(cls.CHANCE, actions, None, chance)

    @classmethod
    def new_terminal(cls, payoffs):
        return cls(cls.TERMINAL, (), payoffs, None)

    def is_chance(self):
        return self.player == self.CHANCE

    def is_terminal(self):
        return self.player == self.TERMINAL


@attr.s(slots=True, cmp=False, frozen=True)
class Situation:
    """
    One gae history and associated structures: observations, active player and actions, state.
    """
    history = attr.ib(type=tuple)
    history_idx = attr.ib(type=tuple)
    active = attr.ib()  #type=ActivePlayer)
    observations = attr.ib(type=tuple)
    state = attr.ib(type=Any)
    game = attr.ib(type='Game')

    def __len__(self):
        return len(self.history)

    @property
    def player(self):
        return self.active.player

    @property
    def actions(self):
        return self.active.actions

    @property
    def chance(self):
        return self.active.chance

    @property
    def payoff(self):
        return self.active.payoff

    def is_terminal(self):
        return self.active.is_terminal()

    def is_chance(self):
        return self.active.is_chance()


@attr.s(slots=True, cmp=False, frozen=True)
class SituationUpdate:
    active = attr.ib()
    state = attr.ib(default=None)
    observations = attr.ib(default=None)


class Game:
    """
    Base class for game instances.

    Every descendant must have an attribute `self.players`.
    Players are numbered `0 .. players - 1`.
    """

    def __init__(self):
        self.players = None

    def initial_state(self) -> Tuple[Any, ActivePlayer]:
        """
        Return the initial internal state and active player.

        Note that the initial game state must be always the same. If the game start
        depends on chance, use a chance node as the first state.
        """
        raise NotImplementedError

    def update_state(self, state: Situation, action: Any) -> Tuple[Any, ActivePlayer, tuple]:
        """
        Return the updated internal state, active player and per-player observations.

        The observations must have length 0 (no obs for anyone) 
        or (players + 1) (last observation is the public one).
        """
        raise NotImplementedError

    def start(self) -> Situation:
        """
        Create a new initial game state.
        """
        state, active = self.initial_state()
        assert active.player < self.players
        return Situation((), (), active, ((), ) * (self.players + 1), state, self)

    def play(self, hist, action=None, index=None, reuse=False) -> Situation:
        """
        Create and return a new game state by playing given action.

        Action can be given either by value or by index in the available actons (or both).
        """
        if hist.game != self:
            raise ValueError("Playing in wrong game {} (state has {})".format(self, hist.game))
        if (action is None) and (index is None):
            raise ValueError("Pass at least one of `action` and `index`.")
        assert not hist.active.is_terminal()
        if action is None:
            action = hist.active.actions[index]
        if index is None:
            index = hist.active.actions.index(action)
        if hist.active.is_terminal():
            raise ValueError("Playing in terminal state {}".format(hist))
        state, active, obs = self.update_state(hist, action)
        assert active.player < self.players
        assert len(obs) in (0, self.players + 1)
        new_obs = hist.observations
        new_obs = []
        for i in range(self.players + 1):
            o_p, o_a = (), ()
            if i == hist.active.player:
                o_a = (Observation(Observation.OWN_ACTION, action), )  # type: ignore
            if len(obs) > 0 and obs[i] is not None:
                o_p = (Observation(Observation.OBSERVATION, obs[i]), )  # type: ignore
            new_obs.append(hist.observations[i] + o_a + o_p)
        new_obs = tuple(new_obs)
        return Situation(hist.history + (action, ), hist.history_idx + (index, ), active, new_obs,
                         state, self)

    def play_sequence(self, actions=None, *, indexes=None, start: Situation = None,
                      reuse=False) -> List[Situation]:
        """
        Play a sequence of actions, return the last visited state.

        Starts from a given state or `self.start()`. The actions may be given by values or their
        indices in the available action lists.
        """
        if start is None:
            start = self.start()
            reuse = True
        if (actions is None) == (indexes is None):
            raise ValueError("Pass exactly one of `actions` and `indexes`.")
        hist = start
        if actions is not None:
            for a in actions:
                hist = self.play(hist, action=a, reuse=reuse)
                reuse = True
        else:
            for i in indexes:
                hist = self.play(hist, index=i, reuse=reuse)
                reuse = True
        return hist

    def play_strategies(self,
                        strategies,
                        *,
                        rng=None,
                        seed=None,
                        start: Situation = None,
                        reuse=False,
                        stop_when: Callable = None,
                        max_moves: int = None):
        """
        Generate a play based on given strategies (one per player), return the last state.

        Starts from a given state or `self.start()`. Plays until a terminal state is hit, `stop_when(hist)` is True or
        for at most `max_moves` actions (whenever given).
        """
        moves = 0
        rng = get_rng(rng=rng, seed=seed)
        if len(strategies) != self.players:
            raise ValueError("One strategy per player required")
        if start is None:
            start = self.start()
            reuse = True
        hist = start
        while not hist.active.is_terminal():
            if stop_when is not None and stop_when(hist):
                break
            if max_moves is not None and moves >= max_moves:
                break
            if hist.active.is_chance():
                dist = hist.active.chance
            else:
                p = hist.active.player
                dist = strategies[p].strategy(hist)
            assert len(dist) == len(hist.active.actions)
            idx = rng.choice(len(hist.active.actions), p=dist)
            hist = self.play(hist, index=idx, reuse=reuse)
            moves += 1
            reuse = True
        return hist

    def sample_payoff(self, strategies, iterations=100, *, seed=None, rng=None):
        """
        Play the game `iterations` times using `strategies`.
        
        Returns `(mean payoff, payoff variances)` as two numpy arrays.
        """
        rng = get_rng(rng=rng, seed=seed)
        payoffs = [
            self.play_strategies(strategies, rng=rng).active.payoff for i in range(iterations)
        ]
        return (np.mean(payoffs, axis=0), np.var(payoffs, axis=0))

    def __repr__(self):
        return "<{}(...)>".format(self.__class__.__name__)

    def __str__(self):
        "By default, strips the outer '<..>' from `repr(self)`."
        s = repr(self)
        if s.startswith('<') and s.endswith('>'):
            s = s[1:-1]
        return s



from gamegym.algorithms import OutcomeMCCFR, BestResponse
from gamegym.value_learning.valuestore import LinearValueStore
from gamegym.value_learning.valuesgd import SparseSGDLinearValueLearning
from gamegym.algorithms.infosets import InformationSetSampler
from gamegym.games.goofspiel import goofspiel_feaures_cards
from gamegym.games.matrix import matrix_zerosum_features
from gamegym.games import Goofspiel, RockPaperScissors
from gamegym.strategy import UniformStrategy
from gamegym.utils import get_rng
import numpy as np
import pytest


@pytest.mark.slow
def test_goofspiel():
    g = Goofspiel(4, scoring=Goofspiel.Scoring.ZEROSUM)
    mc = OutcomeMCCFR(g, seed=42)
    mc.compute(100)
    vs = LinearValueStore(goofspiel_feaures_cards(g.initial_state()), fix_mean=2.5)
    infosampler = InformationSetSampler(g, mc)
    val = SparseSGDLinearValueLearning(g, goofspiel_feaures_cards, vs, infosampler, seed=43)
    val.compute([mc, mc], 100, 0.1, 0.01)
    print(vs.values)


def test_rps():
    g = RockPaperScissors()
    us = UniformStrategy()
    rng = get_rng(seed=3)
    params = rng.rand(3, 3) - 0.5
    vs = LinearValueStore(params, fix_mean=0.0, regularize_l1=6.0)
    infosampler = InformationSetSampler(g, us)
    val = SparseSGDLinearValueLearning(g, matrix_zerosum_features, vs, infosampler, seed=44)
    val.compute([us, us], 100, 0.1, 0.1)
    val.compute([us, us], 100, 0.01, 0.01)
    val.compute([us, us], 100, 0.001, 0.001)



from ..game import Game, Situation
from ..strategy import Strategy
from ..utils import get_rng
from .expected_features import InfoSetExpectedFeatures
import numpy as np
import scipy.optimize
import scipy as sp

# TODO: Update this to new Game API


class FullSGDZeroSumValueLearning:
    def __init__(self,
                 game,
                 feature_extractor,
                 strategies,
                 expected_features=None,
                 sparse=False,
                 seed=None,
                 rng=None):
        self.rng = get_rng(rng=rng, seed=seed)
        self.game = game
        self.feature_extractor = feature_extractor
        self.sparse = sparse
        if expected_features is None:
            self.expected_features = InfoSetExpectedFeatures(
                game, feature_extractor, strategies, sparse=sparse)
        else:
            self.expected_features = expected_features
        if isinstance(strategies, Strategy):
            self.strategies = (strategies, ) * game.players()
        else:
            self.strategies = tuple(strategies)

        self.p0values = feature_extractor(self.game.initial_state())

    def compute(self):
        pass



from ..strategy import Strategy
from ..game import Situation, Game, ActivePlayer
from .mccfr import OutcomeMCCFR, RegretStrategy
from ..utils import get_rng

import collections
import numpy as np

SupportItem = collections.namedtuple("SupportItem", ["state", "probability"])


class BestResponse(Strategy):
    """
    Compute a best-response strategy by game tree traversal.

    May be very computationaly demanding as it traverses the whole tree on creation.
    `strategies[player]` is ignored and may be e.g. `None`.
    """

    def __init__(self, game: Game, player: int, strategies, max_nodes=1e6):
        assert isinstance(game, Game)
        assert player < game.players
        assert len(strategies) == game.players
        nodes = 0

        # DFS for from state to terminal state or stata of "player"
        def trace(state, probability, supports):
            nonlocal nodes
            nodes += 1
            if nodes > max_nodes:
                raise Exception(
                    "BestResponse traversed more than allowed {} nodes.".format(max_nodes) +
                    "Either increase the limit or consider using approximate best response.")
            # Just to get rid of nodes where distrbution returned pure zero
            if probability == 0.0:
                return 0.0
            p = state.active.player
            if p == player:
                pi = state.observations[player]
                s = supports.setdefault(pi, list())
                s.append(SupportItem(state, probability))
                return 0
            if p == ActivePlayer.TERMINAL:
                return state.active.payoff[player] * probability
            if p == ActivePlayer.CHANCE:
                dist = state.active.chance
            else:
                dist = strategies[p].strategy(state)
            return sum(
                trace(game.play(state, action), pr * probability, supports)
                for pr, action in zip(dist, state.active.actions))

        # DFS from isets to other isets of "player"
        def traverse(iset, support):
            actions = support[0].state.active.actions
            values = []
            br_list = []
            for action in actions:
                new_supports = {}
                value = 0
                best_responses = {}
                br_list.append(best_responses)

                for s in support:
                    value += trace(game.play(s.state, action), s.probability, new_supports)
                for iset2, s in new_supports.items():
                    v, br = traverse(iset2, s)
                    value += v
                    best_responses.update(br)

                values.append(value)

            values = np.array(values)
            mx = values.max()
            is_best = values >= (mx - mx * 0e-6)
            br_result = {}
            bdist = is_best.astype(np.float)
            br_result[iset] = bdist / sum(bdist)
            for br, is_b in zip(br_list, is_best):
                if is_b:
                    br_result.update(br)
            return mx, br_result

        supports = {}
        self.best_responses = {}
        value = trace(game.start(), 1.0, supports)
        for iset2, s in supports.items():
            v, br = traverse(iset2, s)
            value += v
            self.best_responses.update(br)
        self.value = value

    def _strategy(self, observation, n_active, state=None):
        return self.best_responses[observation]


class ApproxBestResponse(Strategy):
    """
    Compute an approximate best-response strategy using MCCFR.

    Uses given number of iterations of OutcomeMCCFR.
    `strategies[player]` is ignored and may be e.g. `None`.
    """

    def __init__(self, game: Game, player: int, strategies, iterations, *, seed=None, rng=None):
        self.rng = get_rng(seed=seed, rng=rng)
        self.player = player
        self.game = game
        self.strategies = list(strategies)
        self.strategies[self.player] = RegretStrategy()
        self.mccfr = OutcomeMCCFR(game, self.strategies, [self.player], rng=self.rng)
        self.mccfr.compute(iterations, burn=0.5)

    def _strategy(self, observation, n_active, state=None):
        return self.strategies[self.player]._strategy(observation, n_active, state)

    def sample_value(self, iterations):
        val = self.game.sample_payoff(self.strategies, iterations, rng=self.rng)[0][self.player]
        return val


def exploitability(game, measured_player, strategy, max_nodes=1e6):
    """
    Exact exploitability of a player strategy in a two player ZERO-SUM game.
    """
    assert measured_player in (0, 1)
    assert isinstance(game, Game)
    assert game.players == 2
    assert isinstance(strategy, Strategy)
    br = BestResponse(game, 1 - measured_player, [strategy, strategy], max_nodes)
    return br.value


def approx_exploitability(game, measured_player, strategy, iterations, seed=None, rng=None):
    """
    Approximate exploitability of a player strategy in a two player ZERO-SUM game.

    Uses given number of iterations of OutcomeMCCFR.
    The value is then taken from a mean of `iterations / 4` plays.
    Note that the "best-response" strategy may be worse than the original if the
    iteration number is too small.
    """
    assert isinstance(game, Game)
    assert game.players == 2
    assert measured_player in (0, 1)
    assert isinstance(strategy, Strategy)
    rng = get_rng(seed=seed, rng=rng)
    br = ApproxBestResponse(game, 1 - measured_player, [strategy, strategy], iterations, rng=rng)
    return br.sample_value(iterations // 2)



import numpy as np
import scipy as sp

from ..algorithms.mccfr import RegretStrategy


class StrategyTrace:
    """
    Record the progress of strategy computation.

    Has several visualisation and projection methods.
    Tracks the behavior in all information sets encountered up to given depth.
    """

    def __init__(self, game, depth, name=""):
        self.game = game
        self.depth = depth
        self.name = name
        # last added strategies
        self.last_strategies = None

        infosets = set()

        def _traverse(state, d):
            if state.is_terminal() or d >= self.depth:
                return
            if not state.is_chance():
                infosets.add((state.player, state.observations[state.player], state.actions))
            for ai, a in enumerate(state.actions):
                _traverse(self.game.play(state, index=ai), d + 1)

        _traverse(self.game.start(), 0)

        # List of infosets measured: [(player, observation, actions)]
        self.infosets = sorted(list(infosets))
        # Iterations, shape (t, )
        self.d_t = []
        # Regrets, shape (t, players)
        self.d_exploitability = []
        # Strategy policies in measured infosets, shape (t, actions)
        # `actions` are: first infoset actions, second infoset actions, ...
        self.d_strat = []
        # Regret matching policies in measured infosets, shape (t, actions)
        # Where regret is undefined uses same policy as strategy
        self.d_regret = []

    def append(self, t, strategies, exploitabilities=None):
        """
        Record strategy profile observation at time `t` (iterations).

        Optionally also records the given exploitabilities.
        """
        assert len(strategies) == self.game.players
        assert (not self.d_t) or self.d_t[-1] < t

        strat_vec = []
        reg_vec = []
        for si, iset, actions in self.infosets:
            s = strategies[si]
            strat = s.strategy(iset, len(actions))
            reg = strat
            if isinstance(s, RegretStrategy):
                entry = s.get_entry(iset, len(actions))
                reg = s.regret_matching(entry[0])
            strat_vec.extend(strat)
            reg_vec.extend(reg)

        self.d_t.append(t)
        self.d_strat.append(np.array(strat_vec))
        self.d_regret.append(np.array(reg_vec))
        if exploitabilities is None:
            exploitabilities = [np.nan, np.nan]
        assert len(exploitabilities) == self.game.players
        self.d_exploitability.append(exploitabilities)
        self.last_strategies = strategies

    def infoset_weight_vec(self, strategies):
        """
        Compute the probability of every tracked infoset being visited (within given depth).

        Returns array of shape `(actions, )`.
        """
        assert len(strategies) == self.game.players
        strat_ps = {i: 0.0 for i in self.infosets}

        def _traverse(state, strat_p, d):
            iset = state.observations[state.player]
            idx = (state.player, iset, state.actions)
            if state.is_terminal() or d >= self.depth:
                return
            if not state.is_chance():
                strat_ps[idx] += strat_p
                s = strategies[state.player]
                strat = s.strategy(state)
            else:
                strat = state.chance
            for ai, a in enumerate(state.actions):
                _traverse(self.game.play(state, index=ai), strat_p * strat[ai], d + 1)

        _traverse(self.game.start(), 1.0, 0)
        strat_vec = []
        for idx in self.infosets:
            strat_vec.extend([strat_ps[idx]] * len(idx[2]))
        return np.array(strat_vec)

    def plot_trace(self, base, color, smooth_regret=5, with_regret=True):
        """
        Create Plotly objects to plot the trace.

        Base must have shape `(2, tot_actions)`.
        Returns Plotly graph objects. Optionally also returns regret traces.
        """
        import plotly.graph_objs as pgo

        d_t = np.array(self.d_t)
        d_text = ["{} its".format(t) for t in d_t]
        d_text_expl = []
        for dti, dt in enumerate(d_text):
            exp = self.d_exploitability[dti]
            if np.isfinite(exp).any():
                exptext = ", ".join("{:.3f}".format(v) for v in exp)
                d_text_expl.append("{} expls {}".format(dt, exptext))
        d_strat = np.array(self.d_strat)
        d_regret = np.array(self.d_regret)
        assert base.shape == (2, d_strat.shape[1])

        proj_strat = np.matmul(d_strat, base.transpose())
        objs = [
            pgo.Scatter(
                x=proj_strat[:, 0],
                y=proj_strat[:, 1],
                name=self.name + " strategy",
                mode="lines+markers",
                legendgroup=str(id(self)),
                hovertext=d_text_expl,
                hoverinfo="text+name",
                hoverlabel=dict(namelength=-1),
                line=dict(color=color, width=1.8, shape='spline'),
                marker=dict(
                    symbol="x",
                    line=dict(width=0.2, color='black'),
                    color=color,
                    size=[0] * (proj_strat.shape[0] - 1) + [10])),
        ]
        if with_regret:
            proj_regret = np.matmul(d_regret, base.transpose())
            if smooth_regret > 1:
                from scipy import signal
                proj_regret = signal.savgol_filter(
                    proj_regret, smooth_regret, min(4, smooth_regret - 1), axis=0)
            objs.extend([
                pgo.Scatter(
                    x=proj_regret[:, 0],
                    y=proj_regret[:, 1],
                    name=self.name + " regret",
                    mode="lines+markers",
                    hovertext=d_text,
                    hoverinfo="text+name",
                    hoverlabel=dict(namelength=-1),
                    visible='legendonly',
                    line=dict(color=color, width=0.6, shape='spline', smoothing=0.3),
                    marker=dict(
                        line=dict(width=0.2, color='black'),
                        symbol="x",
                        color=color,
                        opacity=0.7,
                        size=[0] * (proj_regret.shape[0] - 1) + [10])),
            ])
        return objs

    def plot_exploitabilities(self, color, showlegend=True):
        """
        Create Plotly objects to plot the exploitability.

        Base must have shape `(2, tot_actions)`.
        Returns Plotly graph objects. Optionally also returns regret traces.
        """
        import plotly.graph_objs as pgo

        d_t = np.array(self.d_t)
        d_exp = np.array(self.d_exploitability)

        objs = []
        for p in range(self.game.players):
            objs.append(
                pgo.Scatter(
                    x=d_t,
                    y=d_exp[:, p] * (-1.0 if p == 1 else 1.0),
                    name="{} p{} best response value".format(self.name, p),
                    mode="lines",
                    legendgroup=str(id(self)),
                    hoverinfo="x+y+name",
                    hoverlabel=dict(namelength=-1),
                    showlegend=showlegend,
                    connectgaps=True,
                    line=dict(color=color, width=1.5)))
        return objs

    @classmethod
    def common_PCA_base(cls, traces, with_regret=True, weight_vec=None, dims=2, seed=None):
        """
        Compute a commpon PCA base of the given trace set.
        
        Returns unit L2-norm base of shape `(dims, tot_actions)`.
        """
        import sklearn
        from sklearn.decomposition import PCA

        rows = [t.d_strat for t in traces]
        if with_regret:
            rows.extend([t.d_regret for t in traces])
        pca = PCA(dims, random_state=seed)
        rows = np.concatenate(rows, axis=0)
        if weight_vec is not None:
            rows *= [weight_vec]
        pca.fit(rows)
        base = pca.components_
        assert base.shape[0] == dims
        base = base / sp.linalg.norm(base, axis=1).reshape((-1, 1))
        return base

    @classmethod
    def common_random_base(cls, traces, weight_vec=None, dims=2, seed=None):
        """
        Compute a random base of the given trace set.

        Returns unit L2-norm base of shape `(dims, tot_actions)`.
        Optionally weights the actions by weight_vec
        """
        base = np.random.normal(size=(dims, len(traces[0].d_strat[0])))
        if weight_vec is not None:
            base *= [weight_vec]
        base = base / sp.linalg.norm(base, axis=1).reshape((-1, 1))
        assert base.shape[0] == dims
        return base

    @classmethod
    def common_plot(cls,
                    traces,
                    base,
                    plot_traces=True,
                    plot_exploitabilities=True,
                    palette=None,
                    with_regret=True,
                    title=None):
        """
        Create a Plotly plot with trace and exploitability subplots.
        """
        import plotly

        assert plot_exploitabilities or plot_traces
        if plot_traces and plot_exploitabilities:
            #fig = plotly.tools.make_subplots(insets=[dict(cell=(1,1), l=0.6, w=0.4, h=0.4, b=0.05)], print_grid=False)
            fig = plotly.tools.make_subplots(cols=2, print_grid=False, column_width=[0.6, 0.4])
        else:
            fig = plotly.graph_objs.Figure()
        if plot_traces:
            fig['layout']['yaxis']['scaleanchor'] = 'x'
        #if plot_exploitabilities:
        #    fig['layout']['yaxis2' if plot_traces else 'yaxis']['type'] = 'log'
        if palette is None:
            palette = plotly.colors.DEFAULT_PLOTLY_COLORS
        for ti, t in enumerate(traces):
            if plot_traces:
                fig.add_traces(t.plot_trace(base=base, color=palette[ti], with_regret=with_regret))
            if plot_exploitabilities:
                objs = t.plot_exploitabilities(color=palette[ti], showlegend=not plot_traces)
                if plot_traces:
                    for o in objs:
                        o.xaxis = 'x2'
                        o.yaxis = 'y2'
                fig.add_traces(objs)
        fig['layout'].update(hovermode='x', title=dict(text=title))
        return fig



from ..game import Game, Situation
from ..strategy import Strategy
from ..utils import get_rng
from ..algorithms.infosets import InformationSetSampler
import numpy as np
import scipy.optimize
import scipy as sp

# TODO: Update this to new Game API


class InfoSetExpectedFeatures:
    def __init__(self, game, feature_extractor, strategies, infosetsampler=None, sparse=False):
        """
        """
        self.game = game
        self.feature_extractor = feature_extractor
        self.sparse = sparse
        if infosetsampler is None:
            self.infosetsampler = InformationSetSampler(game, strategies)
        else:
            self.infosetsampler = infosetsampler
        if isinstance(strategies, Strategy):
            self.strategies = (strategies, ) * game.players()
        else:
            self.strategies = tuple(strategies)

        # Zero feature array and feature indices
        self.feature_0 = feature_extractor(self.game.initial_state(), sparse=self.sparse)
        # dict {(player, infoset): Explicit(action)}
        self.info_strategy = {}
        # dict {(player, infoset): {action: (player, infoset)}}
        self.info_next = {}
        # dict {(player, infoset): expected_features}
        self.info_features = {}

        self._construct()

    def _construct(self):
        """
        """
        for player in range(self.game.players()):
            info_dist = self.infosetsampler.info_distribution(player)
            for info in info_dist.values():
                state_dist = self.infosetsampler.state_distribution(player, info)
                state0 = state_dist.values()[0]
                assert state0.player() == player
                ins = {}
                for a in state0.actions():
                    sa = state0.play(a)
                    ins[a] = (sa.player(), sa.player_information(sa.player()))
                self.info_next[(player, info)] = ins
                self.info_strategy[(player, info)] = self.strategies[player].distribution(state0)
                fs = self.feature_0.copy()
                totp = 0.0
                for state, state_p in state_dist.items():
                    for ts, tp in self._terminals_under(state, state_p):
                        totp += tp
                        fs += tp * self.feature_extractor(ts, sparse=self.sparse)
                if totp > 0.0:
                    fs = fs / totp
                self.info_features[(player, info)] = fs

    def _terminals_under(self, state, p0=1.0):
        """
        Iterate over terminal nodes under `state`. Generates `(term_state, p_reach)`
        where the reach corresponds to self.strategies.
        """
        if state.is_terminal():
            yield (state, p0)
        else:
            if state.is_chance():
                dist = state.chance_distribution()
            else:
                dist = self.strategies[state.player()].distribution(state)
            for a, ap in dist.items():
                yield from self._terminals_under(state.play(a), p0 * ap)



from gamegym.algorithms import OutcomeMCCFR, BestResponse
from gamegym.games import DicePoker
import numpy as np
import pytest


def test_dicepoker():
    g = DicePoker()

    s = g.start()
    assert len(s.actions) == 36
    assert s.is_chance()
    s = g.play(s, (2, 5))

    assert s.player == 0
    assert set(s.actions) == set(["continue", "fold", "raise"])

    s1 = g.play(s, "fold")

    assert s1.is_terminal()
    assert s1.active.payoff == (-2, 2)

    s1 = g.play(s, "continue")
    assert s1.player == 1
    assert set(s.actions) == set(["continue", "fold", "raise"])

    s2 = g.play(s1, "fold")
    assert s2.is_terminal()
    assert s2.active.payoff == (2, -2)

    s2 = g.play(s1, "continue")
    assert s2.is_terminal()
    assert s2.active.payoff == (-3, 3)

    s3 = g.play(s1, "raise")
    assert s3.player == 0
    assert set(s3.actions) == set(("continue", "fold"))

    s4 = g.play(s3, "continue")
    assert s4.is_terminal()
    assert s4.active.payoff == (-6, 6)

    s4 = g.play(s3, "fold")
    assert s4.is_terminal()
    assert s4.active.payoff == (-4, 4)

    s1 = g.play(s, "raise")
    assert s1.player == 1
    assert set(s1.actions) == set(("continue", "fold"))

    s2 = g.play(s1, "continue")
    assert s2.is_terminal()
    assert s2.active.payoff == (-6, 6)

    s2 = g.play(s1, "fold")
    assert s2.is_terminal()
    assert s2.active.payoff == (4, -4)


@pytest.mark.slow
def test_dicepoker_mc():

    g = DicePoker()
    mc = OutcomeMCCFR(g, seed=52)
    mc.compute(10000, burn=0.5)

    br0 = BestResponse(g, 0, mc.strategies)
    payoff0 = g.sample_payoff([br0, mc.strategies[1]], 10000, seed=3)[0]

    br1 = BestResponse(g, 1, mc.strategies)
    payoff1 = g.sample_payoff([mc.strategies[0], br1], 10000, seed=4)[0]

    assert payoff0[0] > payoff1[0]
    assert payoff0[0] < 0.4
    assert payoff1[0] > 0.07



from gamegym.algorithms import InformationSetSampler
from gamegym.games import RockPaperScissors
from gamegym.strategy import UniformStrategy, Strategy
from gamegym.game import Situation, Observation
from gamegym.utils import Distribution

import pytest
import numpy as np


def test_infoset():
    g = RockPaperScissors()
    us = UniformStrategy()
    iss = InformationSetSampler(g, [us, us])
    assert iss._player_dist.probs == pytest.approx(np.array([0.5, 0.5]))
    assert iss._infoset_dist[0].probs == pytest.approx(np.array([1.0]))
    assert iss._infoset_dist[1].probs == pytest.approx(np.array([1.0]))
    assert iss._infoset_history_dist[0][()].probs == pytest.approx(np.array([1.0]))
    assert iss._infoset_history_dist[1][()].probs == pytest.approx(np.array([1.0, 1.0, 1.0]) / 3)
    iss.sample_player()
    iss.sample_info()
    assert iss.sample_info(0)[1] == ()
    assert iss.sample_info(1)[1] == ()
    assert isinstance(iss.sample_state()[2], Situation)
    assert isinstance(iss.player_distribution(), Distribution)
    assert isinstance(iss.info_distribution(0), Distribution)
    assert isinstance(iss.state_distribution(0, ()), Distribution)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gamegym",
    version="0.1.2",
    author="Tomáš Gavenčiak",
    author_email="gavento@gmail.com",
    description="Game theory framework, algorithms and game implementations",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gavento/gamegym",
    packages=setuptools.find_packages(exclude=['examples']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    # install with pip install -e .
    install_requires=[
        "numpy>=1.15",
        "attrs>=18.0",
        "tqdm>=4.28",
    ],
    # install with pip install -e .[dev]
    extras_require={
        'dev': [
            'coverage>=4.0',
            'pylint>=2.0',
            'pytest-benchmark>=3.0',
            'pytest>=4.0',
            'flake8>=3.0',
            'yapf',
        ]
    })



from ..game import Game, Situation
from ..strategy import Strategy
from ..utils import get_rng
from ..algorithms.infosets import InformationSetSampler
import numpy as np
import scipy.optimize
import scipy as sp

# TODO: Update this to new Game API


class LPZeroSumValueLearning:
    EPS = 1e-6
    POS_NEG_WEIGHT = 1e-12

    def __init__(self, game, infosetsampler, feature_extractor, strategies, flex_weight=1.0):
        """
        Initializes the value learning instance with given game, information set sampling
        helper, feature extractor (e.g. `matrix_zerosum_features`) and strategy set.
        If only one strategy is given, it is used for both players.
        """
        self.game = game
        self.infosetsampler = infosetsampler
        self.feature_extractor = feature_extractor
        if isinstance(strategies, Strategy):
            strategies = (strategies, strategies)
        self.strategies = tuple(strategies)
        self.flex_weight = flex_weight

        # Zero feature array and feature indices
        self.feature_0 = feature_extractor(self.game.initial_state())
        self.features = tuple(np.ndindex(*self.feature_0.shape))

        # Weights for variables and variable types
        self.var_weights = {}
        self.value_variables = set()
        self.nonneg_variables = set()
        self.flex_variables = set()

        # Result components
        self.opt = None
        self.result = None  # Dict: var -> value
        self.flex_sum = None
        # dict: variable -> coefficient
        self.conds_eq = []
        self.conds_le = []
        # right side constants
        self.conds_eq_right = []
        self.conds_le_right = []

        assert game.players() == 2
        self._construct_lp()

    def compute(self, sparse=True, method='interior-point', options=None):
        """
        Construct the linear program and run the LP solver.

        Return the computed feature values in a feature shape array.
        """
        var_list = sorted(
            set().union(*[set(c.keys()) for c in self.conds_eq],
                        *[set(c.keys()) for c in self.conds_le]),
            key=str)
        var_index = {v: i for i, v in enumerate(var_list)}
        # Weights
        weights = np.zeros(len(var_list))
        for v, w in self.var_weights.items():
            weights[var_index[v]] = w

        # Conditions matrices
        def create_matrix(conds):
            if sparse:
                A = sp.sparse.dok_matrix((len(conds), len(var_list)))
            else:
                A = np.zeros((len(conds), len(var_list)))
            for i, coeffs in enumerate(conds):
                for var, c in coeffs.items():
                    A[i, var_index[var]] = c
            if sparse:
                A = A.tocsr()
            return A

        A_eq = create_matrix(self.conds_eq)
        b_eq = np.array(self.conds_eq_right)
        A_le = create_matrix(self.conds_le)
        b_le = np.array(self.conds_le_right)

        # Variable bounds
        bounds = [(0.0, None) if var in self.nonneg_variables else (None, None)
                  for var in var_list]
        # LP computation and result extraction
        self.opt = scipy.optimize.linprog(
            weights,
            A_ub=A_le,
            b_ub=b_le,
            A_eq=A_eq,
            b_eq=b_eq,
            bounds=bounds,
            method=method,
            options=options)
        if not self.opt.success:
            raise Exception("Computation failed: {}".format(self.opt))
        self.result = {v: self.opt.x[i] for i, v in enumerate(var_list)}
        self.flex_sum = sum(self.result[v] for v in self.flex_variables)

        f = self.feature_0.copy()
        for v in self.features:
            f.__setitem__(v, self.result.get(v, 0.0))
        return f

    def add_condition(self, coeffs, right, le=False, flexvar=None):
        """
        Add a linear constraint where `coeffs` is a dict: `var`: coefficient,
        `right` is the right side, `le` indicates `<=` or `==` constraint.
        If `flexvar` is given, a variable allowing flexibility in both directions
        is added.

        Variables can be any hashable objects (usually tuples). Feature value variables
        (for the first player) are denoted by their feature index, e.g. `(i, j)` for
        matrix features.
        """
        if (flexvar is not None) and (flexvar not in self.var_weights):
            self.var_weights[flexvar] = self.flex_weight
            self.flex_variables.add(flexvar)
            self.nonneg_variables.add(flexvar)
        if le:
            d = dict(coeffs)
            if flexvar is not None:
                d[flexvar] = -1.0
            self.conds_le.append(d)
            self.conds_le_right.append(right)
        else:
            if flexvar is not None:
                d1 = dict(coeffs)
                d1[flexvar] = -1.0
                self.conds_le.append(d1)
                self.conds_le_right.append(right)
                d2 = {v: -c for v, c in coeffs.items()}
                d2[flexvar] = -1.0
                self.conds_le.append(d2)
                self.conds_le_right.append(-right)
            else:
                d = dict(coeffs)
                self.conds_eq.append(d)
                self.conds_eq_right.append(right)

    def _construct_lp(self):
        """
        Construct the LP conditions for features based on all information sets.
        Adds one variable `("val", player, info)` for every information set expected payoff.
        Adds one variable `("flex", player, info, a)` for tightness of the expected payoff
        condition.
        """
        # Add conditions weakly pushing the feature values to 0.0
        for v in self.features:
            pv = ("fpos", v)
            nv = ("fneg", v)
            self.add_condition({pv: 1.0, nv: -1.0, v: -1.0}, 0.0)
            self.nonneg_variables.add(pv)
            self.nonneg_variables.add(nv)
            self.var_weights[pv] = self.POS_NEG_WEIGHT
            self.var_weights[nv] = self.POS_NEG_WEIGHT
        # Add conditions on action values
        for player in range(self.game.players()):
            info_dist = self.infosetsampler.info_distribution(player)
            for info in info_dist.values():
                value_var = ("val", player, info)
                self.value_variables.add(value_var)
                state_dist = self.infosetsampler.state_distribution(player, info)
                state0 = state_dist.values()[0]
                assert state0.player() == player
                strategy = self.strategies[player].distribution(state0)
                for a in state0.actions():
                    flex_var = ("flex", player, info, a)
                    #print("Conditions for player {} in infoset {}, action {}, var {}, terminals:".format(
                    #    player, info, a, value_var))
                    f_w = self.feature_0.copy()
                    for state, state_p in state_dist.items():
                        st_a = state.play(a)
                        for ts, tp in self._terminals_under(st_a, state_p):
                            f_w += tp * self.feature_extractor(ts)
                    if player == 1:
                        f_w = -f_w
                    coeffs = {
                        i: f_w.__getitem__(i)
                        for i in self.features if f_w.__getitem__(i) != 0.0
                    }
                    coeffs[value_var] = -1.0
                    if strategy.probability(a) <= self.EPS:
                        self.add_condition(coeffs, 0.0, le=True, flexvar=flex_var)
                    else:
                        self.add_condition(coeffs, 0.0, le=False, flexvar=flex_var)

    def _terminals_under(self, state, p0=1.0):
        """
        Iterate over terminal nodes under `state`. Generates `(term_state, p_reach)`
        where the reach corresponds to self.strategies.
        """
        if state.is_terminal():
            yield (state, p0)
        else:
            if state.is_chance():
                dist = state.chance_distribution()
            else:
                dist = self.strategies[state.player()].distribution(state)
            for a, ap in dist.items():
                yield from self._terminals_under(state.play(a), p0 * ap)



#!/usr/bin/python3

from ..game import Game, Situation, ActivePlayer
from typing import Any
import numpy as np


class MatrixGame(Game):
    """
    General game specified by a payoff matrix.
    The payoffs are for player `i` are `payoffs[p0, p1, p2, ..., i]`.

    Optionally, you can specify the player actions as
    `[[p1a0, p1a1, ...], [p2a0, ...], ...]` where the labels
    may be anything (numbers or strings are recommended)
    If no action labels are given, numbers are used.
    """

    def __init__(self, payoffs, actions=None):
        self.m = payoffs
        if not isinstance(self.m, np.ndarray):
            self.m = np.array(self.m)
        self.players = len(self.m.shape) - 1
        if self.players != self.m.shape[-1]:
            raise ValueError("Last dim of the payoff matrix must be the number of players.")
        if actions is None:
            self.actions = [list(range(acnt)) for acnt in self.m.shape[:-1]]
        else:
            self.actions = actions
        if tuple(len(i) for i in self.actions) != self.m.shape[:-1]:
            raise ValueError(
                "Mismatch of payoff matrix dims and labels provided: {} vs {}.".format(
                    self.m.shape[:-1], tuple(len(i) for i in self.actions)))

    def initial_state(self):
        """
        Return the initial internal state and active player.
        """
        return ((), ActivePlayer.new_player(0, self.actions[0]))

    def update_state(self, hist, action):
        """
        Return the updated internal state, active player and per-player observations.
        """
        # next active player
        p = len(hist.history) + 1
        idx = self.actions[p - 1].index(action)
        if p >= self.players:
            assert p == self.players
            return ((), ActivePlayer.new_terminal(self.m[hist.history_idx + (idx, )]), ())
        return ((), ActivePlayer.new_player(p, self.actions[p]), ())

    def __repr__(self):
        return "<{} {}>".format(self.__class__.__name__, 'x'.join(
            str(x) for x in self.m.shape[:-1]))


class MatrixZeroSumGame(MatrixGame):
    """
    A two-player zero-sum game specified by a payoff matrix.

    The payoffs for player 0 are `payoffs[a0, a1]`, negative for player 1.
    
    Optionally, you can specify the labels for the players as
    `(("p0a0", "p0a1", ...), ("p1a0", ...))`. If no labels are given,
    numbers are used.
    """

    def __init__(self, payoffs, actions=None):
        if not isinstance(payoffs, np.ndarray):
            payoffs = np.array(payoffs, dtype=np.float64)
        super().__init__(np.stack((payoffs, 0.0 - payoffs), axis=-1), actions=actions)


class RockPaperScissors(MatrixZeroSumGame):
    """
    Rock-paper-scissors game with -1,0,1 values.

    Actions are `R(ock)`, `P(aper)` and `S(cissors)`.
    """

    def __init__(self):
        super().__init__([[0, -1, 1], [1, 0, -1], [-1, 1, 0]], (("R", "P", "S"), ("R", "P", "S")))


class GameOfChicken(MatrixGame):
    """
    Game of chicken with customizable values.

    Actions are `D(are)` and `C(hicken)`.
    """

    def __init__(self, win=7, lose=2, both_dare=0, both_chicken=6):
        super().__init__(
            [[[both_dare, both_dare], [win, lose]], [[lose, win], [both_chicken, both_chicken]]],
            (("D", "C"), ("D", "C")))


class PrisonersDilemma(MatrixGame):
    """
    Game of prisoners dilemma with customizable values.

    Actions are `D(efect)` and `C(ooperate)`.
    """

    def __init__(self, win=3, lose=0, both_defect=1, both_cooperate=2):
        super().__init__([[[both_defect, both_defect], [win, lose]],
                          [[lose, win], [both_cooperate, both_cooperate]]], (("D", "C"),
                                                                             ("D", "C")))


class MatchingPennies(MatrixZeroSumGame):
    """
    Game of matchig pennies, the first player is the matcher.

    Actions are `H(eads)` and `T(ails)`.
    """

    def __init__(self, mismatch=1, match_heads=1, match_tails=1):
        super().__init__([[match_heads, -mismatch], [-mismatch, match_tails]], (("H", "T"),
                                                                                ("H", "T")))


def matrix_zerosum_features(hist: Situation, sparse=False):
    assert not sparse
    assert isinstance(hist.game, MatrixGame)
    features = np.zeros(hist.game.m.shape[:-1], dtype=np.float64)
    if hist.active.is_terminal():
        features.__setitem__(hist.history_idx, 1.0)
    return features
